import sys

from mock import patch

# TODO: create a more exhaustive list and move to a different module.
DEFAULT_EXTERNAL_MODULES = [
    # Django
    "django",
    "django.apps",
    "django.test.client",
    "django.conf",
    "django.conf.settings",
    "django.core",
    "django.core.exceptions",
    "django.contrib",
    "django.contrib.auth",
    "django.contrib.auth.decorators",
    "django.db",
    "django.db.models",
    "django.http",
    "django.shortcuts",
    "django.test",
    "django.urls",
    "django.urls.converters",
    "django.utils",
    "django.utils.translation",
    "django.views",
    "django.views.decorators",
    "django.views.decorators.cache",
    "django.views.decorators.clickjacking",
    "django.views.decorators.csrf",
    "django.views.decorators.gzip",
    # Rest Framework
    "rest_framework",
    "rest_framework.decorators",
    "rest_framework.relations",
    "rest_framework.response",
    "rest_framework.serializers",
    "rest_framework.utils",
    "rest_framework.utils.encoders",
    "rest_framework.viewsets",
    # Scientific
    "numpy",
    "pandas",
    # Tools
    "pytest",
    "faker",
    "model_mommy",
    "model_mommy.mommy",
    "factory",
    "factory.declarations",
]


class LightMock:
    """
    A lightweight Mock class.

    It creates attributes and methods on-demand.

    >>> x = LightMock()
    >>> x.foo.bar(42)                                      # doctests: +ELLIPSIS
    <LightMock ...>
    """

    def __init__(self, *args, **kwargs):
        pass

    __method = lambda self, *args, **kwargs: LightMock()
    __call__ = __method
    __getattr__ = __method
    ___name__ = ___qualname__ = "mock"

    # Container interface
    __getitem__ = __method
    __iter__ = lambda self: iter(())

    # Arithmetic operations
    __add__ = __sub__ = __mul__ = __truediv__ = __floordiv__ = __method
    __radd__ = __rsub__ = __rmul__ = __rtruediv__ = __rfloordiv__ = __method

    # Conversions
    __bool__ = lambda self: True

    # Pretend to be a type in some situations
    __mro_entries__ = lambda *args: (object,)
    __getstate__ = lambda self: ()


def mock_save(model, method=LightMock):
    """
    Context manager that mocks the .save() method of a model to prevent it from
    hitting the database.

    Usage:
        .. code-block:: python

            with mock_save(model):
                model.name = "Hello"
                model.save()  # it does not actually touch the db
    """
    return patch.object(model, "save", LightMock)


_stdout = sys.stdout


def mock_modules(*modules):
    """
    Save all given modules that were not imported in sys.modules.
    """

    cls = type("MockItem", (LightMock, type), {})

    if modules == ("auto",):
        modules = DEFAULT_EXTERNAL_MODULES

    def path_hook(path):
        base = path.partition(".")[0]
        if base in modules:
            # It mocks all its way through a mocked module In python import
            # subsystem. This hack makes the implementation much simpler :)
            return cls()
        else:
            raise ImportError

    sys.path_hooks.insert(0, path_hook)

    for mod in modules:
        sys.modules.setdefault(mod, LightMock())


def assume_unique(form=None):
    """
    Context manager that suppress checks of uniqueness during model validation.

    Usage:
        .. code-block:: python

            with assume_unique(model):
                model.slug = "repeated-slug"
                model.full_clean()  # prevents touching the db on uniqueness checks
    """
    from django import forms

    if form is None:
        form = forms.ModelForm
    return patch.object(form, "validate_unique", no_op)


#
# Auxiliary functions
#
no_op = lambda *args, **kwargs: None
cte = lambda cte: lambda *args, **kwargs: cte


def raise_exception(exception):
    """
    Return a function that raises the given exception when called.
    """
    if exception is not None:

        def action(*args, **kwargs):
            raise exception

    else:
        action = no_op
    return action



from .installed_apps import InstalledAppsConf


class MiddlewareConf(InstalledAppsConf):
    """
    Configure middleware classes.
    """

    with_middleware = InstalledAppsConf.with_app

    def get_middleware(self):
        apps = self.INSTALLED_APPS

        def add(middleware):
            if middleware.partition(".middleware")[0] in apps:
                return [middleware]
            else:
                return []

        return [
            "django.middleware.security.SecurityMiddleware",
            *add("django.contrib.sessions.middleware.SessionMiddleware"),
            "django.middleware.common.CommonMiddleware",
            "django.middleware.csrf.CsrfViewMiddleware",
            *add("django.contrib.auth.middleware.AuthenticationMiddleware"),
            *add("django.contrib.messages.middleware.MessageMiddleware"),
            "django.middleware.clickjacking.XFrameOptionsMiddleware",
        ]



from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.http import (
    HttpResponsePermanentRedirect,
    HttpResponseRedirect,
    HttpResponseGone,
    Http404,
)
from django.urls import reverse
from django.utils.translation import ugettext as _

from boogie.views.base import View, log
from boogie.views.mixins import TemplateMixin


class RedirectView(View):
    """
    Provide a redirect on any GET request.
    """

    permanent = False
    url = None
    pattern_name = None
    query_string = False

    def get_redirect_url(self, request, **kwargs):
        """
        Return the URL redirect to. Keyword arguments from the URL pattern
        match generating the redirect request are provided as kwargs to this
        method.
        """
        if self.url:
            url = self.url.format(**kwargs)
        elif self.pattern_name:
            url = reverse(self.pattern_name, kwargs=kwargs)
        else:
            return None

        query_args = request.META.get("QUERY_STRING", "")
        if query_args and self.query_string:
            url = f"{url}?{query_args}"
        return url

    def get(self, request, *args, **kwargs):
        url = self.get_redirect_url(*args, **kwargs)
        if url:
            if self.permanent:
                return HttpResponsePermanentRedirect(url)
            else:
                return HttpResponseRedirect(url)
        else:
            log.warning(
                "Gone: %s", request.path, extra={"status_code": 410, "request": request}
            )
            return HttpResponseGone()

    head = post = options = delete = put = patch = get


class DetailView(TemplateMixin, View):
    """
    Provide the ability to retrieve a single object for further manipulation.
    """

    model = None
    queryset = None
    slug_field = "slug"
    context_object_name = None
    slug_url_kwarg = "slug"
    pk_url_kwarg = "pk"
    query_pk_and_slug = False

    def get_queryset(self, request, **kwargs):
        """
        Return the `QuerySet` that will be used to look up the object.

        This method is called by the default implementation of get_object() and
        may not be called if get_object() is overridden.
        """
        if self.queryset is None:
            if self.model:
                return self.model._default_manager.all()
            else:
                raise ImproperlyConfigured(
                    "%(cls)s is missing a QuerySet. Define "
                    "%(cls)s.model, %(cls)s.queryset, or override "
                    "%(cls)s.get_queryset()." % {"cls": self.__class__.__name__}
                )
        return self.queryset.all()

    def get_object(self, request, **kwargs):
        """
        Return the object the view is displaying.

        Require `self.queryset` and a `pk` or `slug` argument in the URLconf.
        Subclasses can override this to return any object.
        """
        # Use a custom queryset if provided; this is required for subclasses
        # like DateDetailView
        queryset = self.get_queryset(request, **kwargs)
        if queryset is None:
            queryset = self.get_queryset()

        # Next, try looking up by primary key.
        pk = kwargs.get(self.pk_url_kwarg)
        slug = kwargs.get(self.slug_url_kwarg)
        if pk is not None:
            queryset = queryset.filter(pk=pk)

        # Next, try looking up by slug.
        if slug is not None and (pk is None or self.query_pk_and_slug):
            slug_field = self.get_slug_field()
            queryset = queryset.filter(**{slug_field: slug})

        # If none of those are defined, it's an error.
        if pk is None and slug is None:
            raise AttributeError(
                "Generic detail view %s must be called with "
                "either an object pk or a slug." % self.__class__.__name__
            )

        try:
            # Get the single item from the filtered queryset
            obj = queryset.get()
        except queryset.model.DoesNotExist:
            raise Http404(
                _("No %(verbose_name)s found matching the query")
                % {"verbose_name": queryset.model._meta.verbose_name}
            )
        return obj

    def get_slug_field(self):
        """Get the name of a slug field to be used to look up by slug."""
        return self.slug_field

    def get_context_object_name(self, obj):
        """Get the name to use for the object."""
        if self.context_object_name:
            return self.context_object_name
        elif isinstance(obj, models.Model):
            return obj._meta.model_name
        else:
            return None

    def get_context_data(self, **kwargs):
        """Insert the single object into the context dict."""
        context = {}
        if self.object:
            context["object"] = self.object
            context_object_name = self.get_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        context.update(kwargs)
        return super().get_context_data(**context)

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)
        return self.render(context)

    template_name_field = None
    template_name_suffix = "_detail"

    def get_template_names(self, request, **kwargs):
        """
        Return a list of template names to be used for the request. May not be
        called if render_to_response() is overridden. Return the following list:

        * the value of ``template_name`` on the view (if provided)
        * the contents of the ``template_name_field`` field on the
          object instance that the view is operating upon (if available)
        * ``<app_label>/<model_name><template_name_suffix>.html``
        """
        try:
            names = super().get_template_names(request, **kwargs)
        except ImproperlyConfigured:
            # If template_name isn't specified, it's not a problem --
            # # we just start with an empty list.
            names = [*self.get_object_templates(request, **kwargs)]

            # # If self.template_name_field is set, grab the value of the field
            # # of that name from the object; this is the most specific template
            # # name, if given.
            # if request.object and self.template_name_field:
            #     name = getattr(self.object, self.template_name_field, None)
            #     if name:
            #         names.insert(0, name)
            #
            # # The least-specific option is the default <app>/<model>_detail.html;
            # # only use this if the object in question is a model.
            # if isinstance(self.object, models.Model):
            #     object_meta = self.object._meta
            #     names.append("%s/%s%s.html" % (
            #         object_meta.app_label,
            #         object_meta.model_name,
            #         self.template_name_suffix
            #     ))
            # elif hasattr(self, 'model') and self.model is not None and issubclass(self.model, models.Model):
            #     names.append("%s/%s%s.html" % (
            #         self.model._meta.app_label,
            #         self.model._meta.model_name,
            #         self.template_name_suffix
            #     ))
            #
            # # If we still haven't managed to find any template names, we should
            # # re-raise the ImproperlyConfigured to alert the user.
            if not names:
                raise

        return names

    def get_object_templates(self, request, **kwargs):
        return []



from ..django_conf import Conf


class CeleryConf(Conf):
    CELERY_BROKER_URL = "redis://localhost:6379/0"
    CELERY_ACCEPT_CONTENT = ["json"]
    CELERY_TASK_SERIALIZER = "json"
    CELERY_RESULT_SERIALIZER = "json"



from boogie.router import Router

from . import models

app_name = "fragments"
urlpatterns = Router(
    models={"fragment": models.Fragment}, lookup_field={"fragment": "ref"}
)


@urlpatterns.route("<model:fragment>")
def index(request, fragment):
    return fragment.render(request)



import sidekick as sk

from django import http

BOOGIE_VIEW_MIDDLEWARES = {}
NOTGIVEN = object()


def groups_required(groups, **kwargs):
    """
    Return a middleware that validates if the current user belongs to all of
    the provided groups.
    """
    kwargs.setdefault("message", "Insufficient permissions")
    return access_control(lambda r: r.user.has_groups(groups), **kwargs)


def perms_required(perms, *, fn=None, **kwargs):
    """
    Return a middleware that validates if the current user has the given
    permissions.
    """
    kwargs.setdefault("message", "Insufficient permissions")
    if fn:
        return access_control(lambda r: r.user.has_perms(perms, fn(r)), **kwargs)
    else:
        return access_control(lambda r: r.user.has_perms(perms), **kwargs)


def staff_required(**kwargs):
    """
    Return a middleware that validates if the current user is staff.
    """
    kwargs.setdefault("message", "Insufficient permissions")
    return access_control(lambda r: r.user.is_staff, **kwargs)


def login_required(login_url=None, **kwargs):
    """
    Return a middleware that validates if the current user is authenticated.
    """
    if login_url is False:
        kwargs.setdefault("message", "Requires login")
        return access_control(lambda r: r.user.is_authenticated, **kwargs)

    from django.contrib.auth.decorators import login_required

    def middleware(get_response):
        return login_required(login_url=login_url, **kwargs)(get_response)

    return middleware


def check_rule(rule, fn: callable = None, **kwargs):
    """
    Return a middleware that validates if the current user respects the
    given rule.
    """
    import rules

    if fn is None:
        test = lambda r: rules.test_rule(rule, r.user)
    else:
        test = lambda r: rules.test_rule(rule, r.user, fn(r))
    return access_control(test, **kwargs)


def access_control(
    grant_access, response_type=http.HttpResponseForbidden, message="Not allowed"
):
    def middleware(get_response):
        def handler(request):
            if grant_access(request):
                return get_response(request)
            return response_type(message)

        return handler

    return middleware


def unpoly_middleware(get_response):
    """
    Add unpoly headers to request and response objects.
    """

    def handler(request):
        # Read request headers and creates the request.up object
        value = request.META.get
        request.up = sk.record(
            location=value("X-Up-Location"),
            target=value("X-Up-Target"),
            fail_target=value("X-Up-Fail-Target"),
            validate=value("X-Up-Validate"),
        )

        # Handle response and extract title from the context
        response = get_response(request)
        if hasattr(response, "context"):
            title = request.context.get("title")
            if title:
                response.META.setdefault("X-Up-Title", title)

        return response

    return handler


BOOGIE_VIEW_MIDDLEWARES.update(
    {
        "staff_required": staff_required(),
        "login_required": login_required(),
        "unpoly": unpoly_middleware,
    }
)



from collections import deque

from django.utils.translation import ugettext_lazy as _
from hyperpython import Blob
from hyperpython.fragment import fragment as hp_fragment, FragmentNotFound

from .models import Fragment

CACHED_NON_EXISTING = set()
CACHED_FRAGMENTS = {}
CACHE_FRAGMENT_ORDER = deque()
CACHE_MAX_SIZE = 1024
MISSING_FRAGMENT_MSG = _('<div class="error">Missing "{name}" fragment</div>')


def fragment(ref, *, request=None, raises=False, **kwargs):  # noqa: C901
    """
    Return a fragment instance with the given name.
    """
    tried_hyperpython = False

    # Fetch from cache
    try:
        func = CACHED_FRAGMENTS[ref]
    except KeyError:
        pass
    else:
        return func(request=request, **kwargs)

    # If not in cache, test non-existing cache to fast-track to the hyperpython
    # implementation (which usually does not hit the db).
    if ref in CACHED_NON_EXISTING:
        try:
            return hp_fragment(ref, request=request, **kwargs)
        except FragmentNotFound:
            tried_hyperpython = True

    # Now is time to load fragment from the database
    try:
        obj = Fragment.objects.get(ref=ref)
    except Fragment.DoesNotExist:
        CACHED_NON_EXISTING.add(ref)
    else:
        CACHED_FRAGMENTS[ref] = lambda **kwargs: obj.render(**kwargs)
        CACHE_FRAGMENT_ORDER.append(ref)
        while len(CACHED_FRAGMENTS) > CACHE_MAX_SIZE:
            del CACHED_FRAGMENTS[CACHE_FRAGMENT_ORDER.popleft()]

        return obj.render(request=request, **kwargs)

    # Try hyperpython again, if haven't done before
    if not tried_hyperpython:
        try:
            return hp_fragment(ref, request=request, **kwargs)
        except FragmentNotFound:
            if raises:
                raise
    elif raises:
        raise FragmentNotFound(ref)
    return Blob(MISSING_FRAGMENT_MSG.format(name=ref))


def invalidate_cache(name=None):
    """
    Invalidate cache for the given fragment.

    If called without arguments, invalidate cache for all fragments.
    """
    if name is None:
        CACHED_NON_EXISTING.clear()
        CACHED_FRAGMENTS.clear()
    else:
        CACHED_NON_EXISTING.discard(name)
        CACHED_FRAGMENTS.pop(name, None)
        if name in CACHE_FRAGMENT_ORDER:
            CACHE_FRAGMENT_ORDER.remove(name)


fragment.register = hp_fragment.register



import functools
import inspect
import re
from collections import defaultdict

import sidekick as sk
from django.contrib.auth.decorators import (
    login_required,
    permission_required,
    user_passes_test,
)
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.http import HttpResponse, HttpResponseForbidden, Http404
from django.shortcuts import render, redirect
from django.urls import path
from django.views.decorators.cache import never_cache, cache_control
from django.views.decorators.clickjacking import (
    xframe_options_exempt,
    xframe_options_deny,
    xframe_options_sameorigin,
)
from django.views.decorators.csrf import (
    csrf_exempt,
    ensure_csrf_cookie,
    requires_csrf_token,
    csrf_protect,
)
from django.views.decorators.gzip import gzip_page

from .paths import register_model_converter, get_lookup_type
from ..render import render_response


class Cte:
    def __init__(self, cte):
        self.value = cte

    def __call__(self, *args, **kwargs):
        return self.value

    def __repr__(self):
        return "Cte(%r)" % self.value


class ModelLookupMixin:
    """
    Common initialization patterns for Route and Router.
    """

    def __init__(self, models, lookup_field, lookup_type):
        # Normalize models
        self.models = dict(models or ())
        if isinstance(lookup_field, defaultdict):
            self.lookup_field = lookup_field.copy()
        else:
            self.lookup_field = to_default_dict(lookup_field or {}, "pk")
        if isinstance(lookup_type, defaultdict):
            self.lookup_type = lookup_type.copy()
        else:
            self.lookup_type = to_default_dict(lookup_type or {}, "str")


class Route(ModelLookupMixin):
    """
    A route is a combination of a Django view + some url pattern.
    """

    _default_redirect = staticmethod(
        user_passes_test(lambda x: False)(lambda request: None)
    )

    def __init__(
        self,
        path,
        func,
        name=None,
        method=None,
        template=None,
        login=False,
        perms=None,
        staff=False,
        cache=None,
        gzip=False,
        xframe=None,
        csrf=None,
        models=None,
        lookup_field="pk",
        lookup_type=None,
        decorators=(),
        login_url=None,
        perms_policy="default",
        missing_object_policy="default",
    ):
        super().__init__(models, lookup_field, lookup_type)
        self.path = path
        self.function = func
        self.name = normalize_name(name, func)
        self.method = method
        self.template = template
        self.login = login
        self.perms = perms
        self.staff = staff
        self.cache = cache
        self.gzip = gzip
        self.xframe = xframe
        self.csrf = csrf
        self.decorators = tuple(decorators)
        self.login_url = login_url
        self.perms_policy = perms_policy
        self.missing_object_policy = missing_object_policy

        if login or perms or staff:
            self._test_user = True
            self._register_perms(perms)
        else:
            self._test_user = False

    def _register_perms(self, perms):
        # Perms needs special handling: we only pass simple permissions that
        # do not involve access to some object when applying the decorator.
        # Complex perms must be handled inside the view function and cannot
        # use Django's standard @permission_required() mechanism.
        if isinstance(perms, str):
            perms = [perms]
        complex, simple = sk.partition_at(lambda x: ":" in x, perms or ())
        self._simple_perms = tuple(simple)

        # Create a tuple of (obj, [list of perms]) for
        complex_perms = []
        for full_perm in complex:
            perm, _, obj = full_perm.partition(":")

            # Check if any object permissions were defined inconsistently
            if obj and obj not in self.models:
                raise ImproperlyConfigured(
                    f"invalid permission detected: {obj!r} object is required in \n"
                    f"{full_perm!r} permission, but it is not registered in this "
                    f"route."
                )
            complex_perms.append((perm, obj))

        self._complex_perms = tuple(complex_perms)

    def view_function(self):
        """
        Return a function that respects Django's view function contract.
        """

        function = as_request_function(self.function)
        decorators = ["cache", "gzip", "xframe", "csrf", "decorators"]
        kwargs = {attr: getattr(self, attr) for attr in decorators}

        # Creates the default view function.
        @apply_decorators(**kwargs)
        def view_function(request, **kwargs):
            try:
                self.prepare_arguments(request, kwargs)
                result = function(request, **kwargs)
                return self.prepare_response(result, request)
            except HttpExceptional as exc:
                return exc.get_response()

        return view_function

    def path_handler(self, prefix="", path_prefix=""):
        """
        Returns a django.urls.path (or re_path) object that handles the given
        route.
        """
        path_ = self.compatible_path(path_prefix)
        return path(path_, self.view_function(), name=prefix + self.name)

    def prepare_response(self, result, request):
        """
        Return a Django response object from the result of the supplied view
        function.
        """
        # Regular http responses
        if isinstance(result, HttpResponse):
            return result

        # Template context
        elif isinstance(result, dict) and self.template is not None:
            return render(request, self.template, result)

        # Other types
        return render_response(result)

    def prepare_arguments(self, request, args):
        """
        Transforms the dictionary of input arguments inplace.

        This step transforms the view arguments before passing them to the view
        function.
        """
        models = self.models
        if self.missing_object_policy is not None:
            for arg, obj in args.items():
                if obj is None and arg in models:
                    self.handle_object_not_found(request)

        if self._test_user:
            user = request.user
            if (
                self.login
                and not user.is_authenticated
                or self.staff
                and not user.is_staff
                or not user.has_perms(self._simple_perms)
            ):
                self.handle_permission_denied(request)

            for perm, obj_name in self._complex_perms:
                obj = args[obj_name]
                if not user.has_perm(perm, obj):
                    self.handle_permission_denied(request)

    def compatible_path(self, path_prefix=""):
        """
        Convert a Boogie-style path specification to a valid Django path.
        """
        path = path_prefix + self.path

        for name, model in self.models.items():
            if isinstance(model, type) and issubclass(model, Model):
                queryset = None
            else:
                model, queryset = model.model, model

            part = f"<model:{name}>"
            if part in path:
                lookup_field = self.lookup_field[name]
                lookup_type = self.lookup_type[name]
                if lookup_type is None:
                    lookup_type = get_lookup_type(None, model, lookup_field)
                converter = get_converter(model, lookup_field, lookup_type, queryset)
                path = path.replace(part, f"<{converter}:{name}>")

        # Find unregistered models
        if "<model:" in path:
            m = re.search(r"<model:[^>]*>", path)
            part = path[m.start() : m.end()]
            name = part[7:-1]
            raise ImproperlyConfigured(
                f"Could not find a model for {part}. Please pass the correct "
                f"model for {name} in the models dictionary of the route."
            )
        return path

    def handle_object_not_found(self, request):
        """
        Decide what to do if object passed to some parameter of the view
        function is not found on the database.
        """
        policy = self.missing_object_policy
        if policy in ("default", 404):
            raise Http404
        elif policy in ("forbidden", 403):
            raise PermissionError
        elif policy in ("redirect", 302, 303):
            raise HttpExceptional(self._default_redirect(request))
        elif isinstance(policy, int):
            raise HttpExceptional(status_code=policy)
        else:
            raise ImproperlyConfigured("invalid policy")

    def handle_permission_denied(self, request):
        """
        Decide what to do if user does not have permission to see the page.
        """
        policy = self.perms_policy
        if policy in ("default", "redirect", 302, 303):
            raise HttpExceptional(self._default_redirect(request))
        elif policy in ("forbidden", 403):
            raise PermissionError
        elif policy in ("hide", 404):
            raise Http404
        elif isinstance(policy, int):
            raise HttpExceptional(status_code=policy)
        else:
            raise ImproperlyConfigured("invalid permission policy")


#
# Auxiliary functions
#
def as_request_function(function):
    """
    Inspect to check if function receives a request as first parameter and
    return either the original function (if it expects a request) or a
    transformed function that omits the first parameter.
    """
    spec = inspect.getfullargspec(function)

    if spec.args and spec.args[0] == "request":
        return function

    @functools.wraps(function)
    def request_function(request, *args, **kwargs):
        return function(*args, **kwargs)

    return request_function


def apply_decorators(  # noqa: C901
    view=None,
    login=False,
    staff=False,
    perms=None,
    cache=None,
    gzip=False,
    xframe=None,
    csrf=None,
    decorators=(),
):
    """
    Apply decorators to view function. Can also be used as a decorator.
    """

    if view is None:
        kwargs = locals()
        kwargs.pop("view")
        return lambda view: apply_decorators(view, **kwargs)

    # Cache control
    if cache is False:
        view = never_cache(view)
    elif cache is not None:
        view = cache_control(**cache)(view)

    # Permissions
    # (We keep the implementation here, but those options are not handled by
    #  this decorator anymore).
    if login:
        view = login_required(view)
    if perms:
        view = permission_required(perms)(view)
    if staff:
        view = staff_required(view)

    # Compression
    if gzip:
        view = gzip_page(view)

    # Security
    if xframe is False:
        view = xframe_options_exempt(view)
    elif xframe == "deny":
        view = xframe_options_deny(view)
    elif xframe == "sameorigin":
        view = xframe_options_sameorigin(view)
    if csrf is False:
        view = csrf_exempt(view)
    elif csrf == "cookie":
        view = ensure_csrf_cookie(view)
    elif csrf == "token":
        view = requires_csrf_token(view)
    elif csrf is True:
        view = csrf_protect(view)

    # Apply final decorators
    for decorator in reversed(decorators):
        view = decorator(view)
    return view


def staff_required(view):
    @functools.wraps(view)
    def decorated(request, **kwargs):
        if not request.user.is_staff:
            return HttpResponseForbidden()
        return view(request, **kwargs)

    return decorated


def normalize_lookup(field, name=None):
    if isinstance(field, str):
        return field
    return field.get(name)


def normalize_name(name, function=None):
    if name:
        return name
    return function.__name__.replace("_", "-")


def to_default_dict(value, default=None):
    if isinstance(value, defaultdict):
        return value.copy()
    elif isinstance(value, dict):
        if default is not None:
            data = defaultdict(Cte(default))
            data.update(value)
            return data
        else:
            return dict(value)
    else:
        return defaultdict(Cte(value))


@functools.lru_cache(maxsize=256)
def get_converter(model, lookup_field, lookup_type, queryset):
    if queryset is None:
        name = f"${model.__name__}.{lookup_field}-{lookup_type}"
    else:
        name = f"${model.__name__}.{lookup_field}-{lookup_type}-{id(queryset)}"
    register_model_converter(model, name, lookup_field, lookup_type, queryset)
    return name


#
# Exceptional handling of events
#
class HttpExceptional(Exception):
    """
    Exception raised
    """

    def __init__(self, msg=None, status_code=500, cls=HttpResponse, **kwargs):
        super().__init__(msg, status_code, cls, kwargs)

    def get_response(self):
        msg, status_code, cls, kwargs = self.args
        if isinstance(msg, HttpResponse):
            return msg
        return cls(msg, status=status_code, **kwargs)


class HttpRedirect(HttpExceptional):
    def __init__(self, to, **kwargs):
        super(Exception, self).__init__(to, kwargs)

    def get_response(self):
        to, kwargs = self.args
        return redirect(to, **kwargs)



class SocialAuthConf:
    # Social authentication
    SOCIAL_AUTH_LOGIN_REDIRECT_URL = "/"
    SOCIAL_AUTH_LOGIN_URL = "/auth/login/"

    # OAUTH keys (not working yet)
    SOCIAL_AUTH_GITHUB_KEY = ""
    SOCIAL_AUTH_GITHUB_SECRET = ""

    SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = ""
    SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = ""
    SOCIAL_AUTH_GOOGLE_OAUTH2_SCOPE = ""

    SOCIAL_AUTH_FACEBOOK_KEY = "1085127354890672"
    SOCIAL_AUTH_FACEBOOK_SECRET = "9f6ba5c8721172acaf25a733c4c81a99"
    SOCIAL_AUTH_FACEBOOK_SCOPE = ["email"]

    SOCIAL_AUTH_TWITTER_KEY = "JrIVXbXZguPeUsnqcjtbQEWXH"
    SOCIAL_AUTH_TWITTER_SECRET = "vMZjdO7DsUV8mVo46smQK2SHyhCxnXyc24gxH6J6cH08anWqHA"



# Generated by Django 2.1.7 on 2019-03-10 22:15

import boogie.apps.fragments.enums
import boogie.fields.enum_field
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Fragment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "ref",
                    models.CharField(
                        db_index=True,
                        help_text="Unique identifier for fragment name",
                        max_length=100,
                        unique=True,
                        verbose_name="Identifier",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        blank=True,
                        help_text="Optional description that helps humans identify the content and role of the fragment.",
                        max_length=100,
                    ),
                ),
                (
                    "format",
                    boogie.fields.enum_field.EnumField(
                        boogie.apps.fragments.enums.Format,
                        help_text="Defines how content is interpreted.",
                        verbose_name="Format",
                    ),
                ),
                (
                    "content",
                    models.TextField(
                        blank=True,
                        help_text="Raw fragment content in HTML or Markdown",
                        verbose_name="content",
                    ),
                ),
                ("editable", models.BooleanField(default=True, editable=False)),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="FragmentLock",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "fragment",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="lock_ref",
                        to="fragments.Fragment",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
    ]



import os
import sys

from invoke import task

python = sys.executable


@task
def run(ctx, no_toolbar=False):
    """
    Run development server
    """
    env = {}
    if no_toolbar:
        env["DISABLE_DJANGO_DEBUG_TOOLBAR"] = "true"
    else:
        manage(ctx, "runserver 0.0.0.0:8000", env=env)


@task
def gunicorn(ctx, debug=None, environment="production", port=8000, workers=4):
    """
    Run application using gunicorn for production deploys.

    It assumes that static media is served by a reverse proxy.
    """

    from gunicorn.app.wsgiapp import run as run_gunicorn

    directory = ""

    env = {
        "DISABLE_DJANGO_DEBUG_TOOLBAR": str(not debug),
        "PYTHONPATH": "src",
        "DJANGO_ENVIRONMENT": environment,
    }
    if debug is not None:
        env["DJANGO_DEBUG"] = str(debug).lower()
    os.environ.update(env)
    args = [
        "ej.wsgi",
        "-w",
        str(workers),
        "-b",
        f"0.0.0.0:{port}",
        "--error-logfile=-",
        "--access-logfile=-",
        "--log-level",
        "info",
        f"--pythonpath={directory}/src",
    ]
    sys.argv = ["gunicorn", *args]
    run_gunicorn()


@task
def clean_migrations(ctx, all=False, yes=False):
    """
    Remove all automatically created migrations.
    """
    import re

    auto_migration = re.compile(r"\d{4}_auto_\w+.py")
    all_migration = re.compile(r"\d{4}\w+.py")

    rm_list = []
    for app in os.listdir("src"):
        migrations_path = f"src/{app}/migrations/"
        if os.path.exists(migrations_path):
            migrations = os.listdir(migrations_path)
            if "__pycache__" in migrations:
                migrations.remove("__pycache__")
            if all:
                rm_list.extend(
                    [
                        f"{migrations_path}{f}"
                        for f in migrations
                        if all_migration.fullmatch(f)
                    ]
                )
            elif sorted(migrations) == ["__init__.py", "0001_initial.py"]:
                rm_list.append(f"{migrations_path}/0001_initial.py")
            else:
                rm_list.extend(
                    [
                        f"{migrations_path}/{f}"
                        for f in migrations
                        if auto_migration.fullmatch(f)
                    ]
                )
    remove_files(rm_list, yes)


def remove_files(remove_files, yes):
    print("Listing auto migrations")
    for file in remove_files:
        print(f"* {file}")
    if all:
        print(
            "REMOVING ALL MIGRATIONS IS DANGEROUS AND SHOULD ONLY BE " "USED IN TESTING"
        )
    if yes or input("Remove those files? (y/N)").lower() == "y":
        for file in remove_files:
            os.remove(file)


#
# DB management
#
@task
def db(ctx, migrate_only=False):
    """
    Perform migrations
    """
    if not migrate_only:
        manage(ctx, "makemigrations")
    manage(ctx, "migrate")


@task
def db_reset(ctx):
    """
    Reset data in database and optionally fill with fake data
    """
    ctx.run("rm -f local/db/db.sqlite3")
    manage(ctx, "migrate")


@task
def db_fake(ctx, users=True, conversations=True, admin=True, safe=False, theme=None):
    """
    Adds fake data to the database
    """
    set_theme(theme)
    msg_error = "Release build. No fake data will be created!"

    if safe:
        if os.environ.get("FAKE_DB") == "true":
            print("Creating fake data...")
        else:
            return print(msg_error)
    if users:
        manage(ctx, "createfakeusers", admin=admin)
    if conversations:
        manage(ctx, "createfakeconversations")


#
# Utility functions
#
def manage(ctx, cmd, env=None, **kwargs):
    """
    Call python manage.py in a more robust way
    """
    kwargs = {k.replace("_", "-"): v for k, v in kwargs.items() if v is not False}
    opts = " ".join(f'--{k} {"" if v is True else v}' for k, v in kwargs.items())
    cmd = f"{python} manage.py {cmd} {opts}"
    env = {**os.environ, **(env or {})}
    path = env.get("PYTHONPATH", ":".join(sys.path))
    env.setdefault("PYTHONPATH", f"src:{path}")
    ctx.run(cmd, pty=True, env=env)


set_theme = lambda *args: None



import os

from boogie.configurations.tools import module_exists
from .paths import PathsConf


class TemplatesConf(PathsConf):
    """
    Configure templates.
    """

    def get_templates(self):
        templates = [self.DJANGO_TEMPLATES, self.JINJA_TEMPLATES]
        return [x for x in templates if x]

    #
    # DJANGO TEMPLATES
    #
    def get_django_templates(self):
        return {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": self.DJANGO_TEMPLATES_DIRS,
            "APP_DIRS": True,
            "OPTIONS": {"context_processors": self.DJANGO_CONTEXT_PROCESSORS},
        }

    def get_django_context_processors(self):
        return [
            "django.template.context_processors.debug",
            "django.template.context_processors.request",
            "django.contrib.auth.context_processors.auth",
            "django.contrib.messages.context_processors.messages",
        ]

    #
    # JINJA2 TEMPLATES
    #
    def get_jinja_templates(self):
        options = {}
        env = self.JINJA_ENVIRONMENT
        if env is not None:
            options["environment"] = env
        return {
            "BACKEND": "django.template.backends.jinja2.Jinja2",
            "DIRS": self.JINJA_TEMPLATES_DIRS,
            "APP_DIRS": True,
            "OPTIONS": {"extensions": self.JINJA_EXTENSIONS, **options},
        }

    def get_django_templates_dirs(self):
        return []

    def get_jinja_templates_dirs(self):
        return []

    def get_jinja_extensions(self):
        return ["jinja2.ext.i18n"]

    def get_jinja_environment(self):
        base, _, end = os.environ["DJANGO_SETTINGS_MODULE"].rpartition(".")
        module = base + ".jinja2"
        if module_exists(module):
            return module + ".environment"
        return "boogie.jinja2.boogie_environment"



from boogie.configurations import django_conf
from boogie.configurations.django_conf import DjangoConf


class LoggingConf(django_conf.LoggingConf):
    """
    Same as default config.
    """


class UrlsConf(django_conf.UrlsConf):
    """
    Same as default config.
    """


class PathsConf(django_conf.PathsConf):
    """
    Same as default config.
    """


class TemplatesConf(django_conf.TemplatesConf):
    """
    Same as default config.
    """


class EnvironmentConf(django_conf.EnvironmentConf):
    """
    Same as default config.
    """


class ServicesConf(django_conf.ServicesConf):
    """
    Same as default config.
    """


class SecurityConf(django_conf.SecurityConf):
    """
    Same as default config.
    """


class InstalledAppsConf(django_conf.InstalledAppsConf):
    """
    Install the following apps:
    * Django Rest Framework apps

    Plus:
    * Install Django Debug Toolbar in local environment.
    """

    def get_third_party_apps(self):
        apps = super().get_third_party_apps()
        if self.ENVIRONMENT == "local":
            apps = ["debug_toolbar", *apps, "django_extensions"]
        return apps


class MiddlewareConf(django_conf.MiddlewareConf):
    """
    Changes:
    * Install Django debug toolbar middleware in local enviroment.
    """


class BoogieConf(
    DjangoConf,
    SecurityConf,
    TemplatesConf,
    UrlsConf,
    MiddlewareConf,
    InstalledAppsConf,
    LoggingConf,
    ServicesConf,
    PathsConf,
    EnvironmentConf,
):
    """
    Base configuration for the Boogie stack.

    See each layer of configuration to see the changes introduced by the
    Boogie stack.
    """



from sidekick import import_later as _import_later

from .linear_namespace import linear_namespace, linear_namespace_from_sequence

random_name = _import_later(".random_names:random_name", package=__package__)



"""
Random Phrases
--------------

Generate human-friendly random strings. Phrases are usually of the form
"adjective" + "subjective". We try to use mostly characters from Science and
Science fiction movies.

The main entry point is the :func:`phrase` function, that simply returns a
random phrase. If you want to learn more, read the source :)

.. autofunction :: phrase
"""

import random

from faker import Factory

fake = Factory.create("en-us")

star_wars_characters = [
    "Han Solo",
    "Darth Vader",
    "C3PO",
    "R2D2",
    "Luke Skywalker",
    "Princess Leia",
    "Jabba",
    "Obi Wan",
    "Yoda",
    "Jar Jar Binks",
]
famous_scientists = [
    # Physicists
    "Einstein",
    "Newton",
    "Dirac",
    "Bohr",
    "Rutherford",
    "Heisenberg",
    "Curie",
    "Langevin",
    "Boltzmann",
    # Mathematicians
    "Pythagoras",
    "Peano",
    "Hilbert",
    "Gauss",
    "Galois",
    # Computer science
    "Knuth",
    "Turing",
    "Tim",
    # Biology
    "Darwin",
    "Mendel",
    "Lamarck",
    "Mayr",
    "Dobzhansky",
]
adjective_list = [
    "grumpy",
    "heroic",
    "coward",
    "brave",
    "treacherous",
    "powerful",
    "influential",
]
phrase_groups = []


def is_phase_provider(func):
    phrase_groups.append(func)
    return func


def phrase(maker=None):
    """
    A random easy to memorize phrase.
    """
    maker = random.choice(phrase_groups)
    return maker()


def phrase_lower(maker=None):
    """
    Like phrase, but normalize to lowercase results.
    """
    maker = random.choice(phrase_groups)
    return maker().lower()


def subjective_adjective_phrase(subjectives=None, adjectives=None):
    """
    Return a new subjective-adjective phrase such as "Grumpy Einstein" from a
    list of subjectives and adjectives.
    """
    subjective = random.choice(subjectives or famous_scientists)
    adjective = random.choice(adjectives or adjective_list)
    return "%s %s" % (adjective.title(), subjective)


@is_phase_provider
def random_star_wars_phrase():
    """
    Random phrase based on Star Wars ;-)
    """
    return subjective_adjective_phrase(star_wars_characters)


@is_phase_provider
def random_scientist_phrase():
    """
    Random phrase using important scientists.
    """
    return subjective_adjective_phrase(famous_scientists)


@is_phase_provider
def random_fake_phrase():
    """
    Use fake-factory names.
    """
    return subjective_adjective_phrase([fake.first_name()])


@is_phase_provider
def random_fake_catch():
    """
    Catch phrases
    """
    return fake.catch_phrase()



from .default_fields import (
    AutoSlugField,
    NameField,
    TitleField,
    LongDescriptionField,
    ShortDescriptionField,
)
from .enum_field import EnumField
from .enum_type import IntEnum, Enum



from .environment import EnvironmentConf
from ..descriptors import env


def using_db(db):
    return lambda self: db in self.DATABASES["default"]["ENGINE"]


class DatabaseConf(EnvironmentConf):
    """
    Configure the database.

    See also: https://docs.djangoproject.com/en/2.0/ref/settings/#databases
    """

    DATABASE_FROM_DB_URL = env(
        "sqlite:///local/db/db.sqlite3", type="db_url", name="DJANGO_DB_URL"
    )

    def get_databases(self):
        return {"default": {"TEST": {}, **self.DATABASE_FROM_DB_URL}}

    # Derived inspections
    get_using_sqlite = using_db("sqlite")
    get_using_postgresql = using_db("postgresql")
    get_using_mysql = using_db("mysql")



from .descriptors import env, env_settings, env_default
from .base import Conf, save_configuration
from .django_conf import (
    DjangoConf,
    EnvironmentConf,
    InstalledAppsConf,
    MiddlewareConf,
    TemplatesConf,
    SecurityConf,
    PathsConf,
    UrlsConf,
    ServicesConf,
    LoggingConf,
    locales,
)



from ..base import Conf


class UserenaConf(Conf):
    USERENA_ACTIVATION_REQUIRED = False
    USERENA_SIGNIN_AFTER_SIGNUP = True
    USERENA_DISABLE_PROFILE_LIST = True
    USERENA_ACTIVATION_DAYS = 7
    USERENA_FORBIDDEN_USERNAMES = (
        "signup",
        "signout",
        "signin",
        "activate",
        "me",
        "password",
        "login",
        "codeschool",
    )
    USERENA_USE_HTTPS = False
    USERENA_REGISTER_PROFILE = False
    USERENA_SIGNIN_REDIRECT_URL = "/"
    USERENA_REDIRECT_ON_SIGNOUT = "/"
    USERENA_PROFILE_LIST_TEMPLATE = "auth/profile-list.jinja2"



from sidekick import import_later as _import

from .enums import Format

fragment = _import(".functions:fragment", package=__package__)
invalidate_cache = _import(".functions:invalidate_cache", package=__package__)



from django.utils.translation import ugettext_lazy as _

from boogie import models
from boogie.fields import EnumField
from . import invalidate_cache
from .enums import Format


class Fragment(models.Model):
    """
    Configurable HTML fragments that can be inserted in pages.
    """

    ref = models.CharField(
        _("Identifier"),
        max_length=100,
        unique=True,
        db_index=True,
        help_text=_("Unique identifier for fragment name"),
    )
    title = models.CharField(
        max_length=100,
        blank=True,
        help_text=_(
            "Optional description that helps humans identify the content and "
            "role of the fragment."
        ),
    )
    format = EnumField(
        Format, _("Format"), help_text=_("Defines how content is interpreted.")
    )
    content = models.TextField(
        _("content"),
        blank=True,
        help_text=_("Raw fragment content in HTML or Markdown"),
    )
    editable = models.BooleanField(default=True, editable=False)

    def __str__(self):
        return self.ref

    def __html__(self):
        return str(self.render())

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        invalidate_cache(self.ref)

    def delete(self, using=None, keep_parents=False):
        super().delete(using, keep_parents)
        invalidate_cache(self.ref)

    def lock(self):
        """
        Prevents fragment from being deleted on the admin.
        """
        FragmentLock.objects.update_or_create(fragment=self)

    def unlock(self):
        """
        Allows fragment being deleted.
        """
        FragmentLock.objects.filter(fragment=self).delete()

    def render(self, request=None, **kwargs):
        """Render element to HTML"""
        return self.format.render(self.content, request, kwargs)


class FragmentLock(models.Model):
    """
    ForeignKey reference that prevents protected fragments from being deleted
    from the database.
    """

    fragment = models.OneToOneField(
        Fragment, on_delete=models.PROTECT, related_name="lock_ref"
    )



import os


def execute(func: callable, *, watch=False, background=False, **kwargs):
    """
    Execute function. It can perform execution on the background or schedule
    execution by watching changes under some given path in the filesystem.

    Args:
        func (callable):
            A callable that receives no arguments.
        watch (bool):
            If True, execute in watch mode. You probably will want to specify
            an optional path argument to determine the watch location.
        background (bool):
            If True, runs in a background thread.

    See Also:
        See :func:`watch` for options on executing in watch mode.
    """

    if watch and background:
        go = lambda: _watch(func, **kwargs)
        return execute(go, background=True)
    elif watch:
        return _watch(func, **kwargs)
    elif background:
        from threading import Thread

        def go():
            try:
                func()
            except KeyboardInterrupt:
                pass

        thread = Thread(target=go, daemon=True)
        thread.start()
    else:
        func()


def watch(  # noqa: C901
    func: callable, path: str = None, *, poll_time=0.5, skip_first=False, name=None
):
    """
    Execute function and re-execute if everytime a file under the given path
    changes.
    """

    import time
    from watchdog.observers import Observer
    from watchdog.events import (
        FileSystemEventHandler,
        FileCreatedEvent,
        FileDeletedEvent,
        FileModifiedEvent,
        FileMovedEvent,
    )

    # Create the dispatch function that throttles execution so func is
    # executed at most every poll_time seconds
    file_event = (FileCreatedEvent, FileDeletedEvent, FileModifiedEvent, FileMovedEvent)
    last = time.time()

    def dispatch(ev):
        nonlocal last

        if (
            ev.src_path.endswith("__")
            or ev.src_path.startswith("__")
            or ev.src_path.startswith("~")
            or ev.src_path.startswith(".")
        ):
            return

        if isinstance(ev, file_event):
            last = start = time.time()
            time.sleep(poll_time)
            if last == start:
                print(f"File modified: {ev.src_path}")
                func()

    # Initialize observer and mokey-match the instance dispatch method
    observer = Observer()
    handler = FileSystemEventHandler()
    handler.dispatch = dispatch
    observer.schedule(handler, path or os.getcwd(), recursive=True)
    observer.start()
    name = name or func.__name__

    # Starts execution loop
    print(f"Running {name} in watch mode.")
    if not skip_first:
        func()
    try:
        while True:
            time.sleep(0.5)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


_watch = watch



from ..base import Conf


class Pipeline(Conf):
    get_pipeline_enabled = lambda self: not self.DEBUG
    get_pipeline_collector_enabled = lambda self: True

    def get_pipeline(self):
        """
        Return pipeline configuration.
        """
        result = {
            "PIPELINE_ENABLED": self.PIPELINE_ENABLED,
            "PIPELINE_COLLECTOR_ENABLED": self.PIPELINE_COLLECTOR_ENABLED,
        }
        for section in ["JAVASCRIPT", "STYLESHEETS"]:
            contents = getattr(self, "PIPELINE_" + section, None)
            if contents is not None:
                result[section] = contents
        return result

    #
    # Integrations
    #
    def get_jinja_extensions(self):
        return [*super().get_jinja_extensions(), "pipeline.jinja2.PipelineExtension"]



from pathlib import Path

from django.core.exceptions import ImproperlyConfigured

from .paths import PathsConf


class LoggingConf(PathsConf):
    """
    Base configurations for logging.
    """

    def get_debug_logger(self):
        return {"handlers": ["console", "file"], "level": "DEBUG", "propagate": True}

    def get_default_logger(self):
        return {"handlers": ["console", "file"], "level": "INFO", "propagate": True}

    def get_logging_console_handler(self):
        return {"level": "DEBUG", "class": "logging.StreamHandler"}

    def get_logging_file_handler(self):
        path = Path(self.LOG_FILE_PATH)
        if path.is_dir():
            raise ImproperlyConfigured(
                "log file is a directory. Please set a correct LOG_FILE_PATH "
                "in your configuration file that points to a log file, instead "
                "of a directory."
            )
        return {"level": "DEBUG", "class": "logging.FileHandler", "filename": path}

    def get_logging(self):  # noqa: N802
        return {
            "version": 1,
            "disable_existing_loggers": False,
            "handlers": {
                "file": self.LOGGING_FILE_HANDLER,
                "console": self.LOGGING_CONSOLE_HANDLER,
            },
            "loggers": {"django": self.DEFAULT_LOGGER, "celery": self.DEFAULT_LOGGER},
        }



from functools import lru_cache

from django.template import engines
from django.utils.translation import ugettext_lazy as _
from hyperpython import Text, Blob
from hyperpython.components import markdown
from sidekick import import_later

from boogie.fields import IntEnum

bleach = import_later("bleach")
django_template = import_later("django.template.backends.django")
jinja2_template = import_later("django.template.backends.jinja2")

RENDERER = {}
renderer = lambda fmt: lambda f: RENDERER.setdefault(fmt, f)


class Format(IntEnum):
    """
    Data format.
    """

    TEXT = 0, _("Text")
    HTML_RAW = 1, _("HTML")
    HTML_TRUSTED = 2, _("Trusted (non-sanitized) HTML data.")
    MARKDOWN = 10, _("Markdown content")
    DJANGO_TEMPLATE = 20, _("A Django template")
    JINJA_TEMPLATE = 21, _("A Jinja template")

    def render(self, data, request, ctx):
        """
        Render a string of text as the given format.
        """
        return RENDERER[self](data, request, ctx)


@renderer(Format.TEXT)
def _(data, request, ctx):
    return Text(data)


@renderer(Format.HTML_RAW)
def _(data, request, ctx):
    return Blob(sanitize_html(data))


@renderer(Format.HTML_TRUSTED)
def _(data, request, ctx):
    return Blob(data)


@renderer(Format.MARKDOWN)
def _(data, request, ctx):
    return markdown(data)


@renderer(Format.DJANGO_TEMPLATE)
def _(data, request, ctx):
    return render_template("django", data, request, ctx)


@renderer(Format.JINJA_TEMPLATE)
def _(data, request, ctx):
    return render_template("jinja2", data, request, ctx)


#
# Auxiliary functions
#
def sanitize_html(html):
    """
    Convert a string of user HTML in safe html.
    """
    return bleach.clean(html, tags=TAG_WHITELIST, attributes=ATTR_WHITELIST)


def render_template(kind, data, request, ctx):
    template = get_template(data, kind)
    rendered = template.render(ctx, request=request)
    return Text(rendered, escape=False)


@lru_cache(1024)
def get_template(data, fmt):
    engine = engines[fmt]

    if fmt == "django":
        return django_template.Template(data, engine)
    elif fmt == "jinja2":
        return jinja2_template.Template(data, engine)
    else:
        raise ValueError(f"invalid template name: {fmt}")


#
# Constants
#

# List of valid tags to pass through the sanitizer
TAG_WHITELIST = bleach.ALLOWED_TAGS + [
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6" "img",
    "div",
    "span",
    "p",
]

# Valid attributes in each tag
ATTR_WHITELIST = bleach.ALLOWED_ATTRIBUTES



import os

from .paths import PathsConf
from ..descriptors import env

# Map country on default configurations.
COUNTRY_DB = {None: {"LANGUAGE_CODE": "en-us", "TIME_ZONE": "Greenwich"}}
COUNTRY_DB[""] = COUNTRY_DB[None]


class LocaleConf(PathsConf):
    """
    Base Django locale configuration.
    """

    USE_I18N = env(True)
    USE_L10N = env(True)
    USE_TZ = env(True)
    COUNTRY = env("", name="{attr}")

    def get_language_code(self, country):
        return country_method("language_code", country)

    def get_time_zone(self, country):
        return country_method("time_zone", country)

    def get_locale_name(self, language_code, country):
        try:
            return country_method("locale_name", country)
        except ValueError:
            if language_code:
                lang, country = language_code.split("-")
                return f"{lang}_{country.upper()}.UTF8"
            else:
                return ""

    def get_locale_paths(self):
        path = os.path.join(self.REPO_DIR, "locale")
        return self.env("LOCALE_PATH", type=list, default=[path])


class NoLocaleConf(PathsConf):
    """
    Base configuration for instances that do not use localization.
    """

    USE_I18N = env(False)
    USE_L10N = env(False)
    USE_TZ = False
    LANGUAGE_CODE = None
    TIME_ZONE = None


def country_method(var, country):
    django_var = "DJANGO_" + var.upper()
    key = var.upper()
    if django_var in os.environ:
        return os.environ[django_var]
    try:
        return COUNTRY_DB[country.upper()][key]
    except KeyError as exc:
        raise ValueError(f"{exc}: unknown for country {country}")


#
# Register countries
#
def country(country: list, language_code, timezone):
    """
    Return a function that creates a LocaleConf subclass specialized for a
    given country.

    This function return a factory rather than the Conf subclass itself to avoid
    the class creation overhead of a big list of countries.
    """
    names = [country] if isinstance(country, str) else country
    for alias in names:
        COUNTRY_DB[alias.upper()] = {
            "LANGUAGE_CODE": language_code,
            "TIME_ZONE": timezone,
        }


# Ursal
country(["Argentina"], "es-ar", "America/Buenos_Aires")
country(["Bolívia"], "es-bo", "America/La_Paz")
country(["Brasil", "Brazil"], "pt-br", "America/Sao_Paulo")
country(["Chile"], "es-cl", "America/Santiago")
country(["Ecuador"], "es-ec", "America/Guayaquil")
country(["Paraguay"], "es-py", "America/Asuncion")
country(["Peru"], "es-pe", "America/Lima")
country(["Uruguay"], "es-uy", "America/Montevideo")
country(["Venezuela"], "es-ve", "America/Caracas")

# Other american countries

# Africa

# Asia

# Europe

# Oceania



import re
import os
from setuptools import setup, find_packages

init = open(os.path.join('src', 'boogie', '__init__.py')).read()
version = re.search(r'__version__ = "(\d+\.\d+.\d+)"', init).group(1)

setup(
    name='django-boogie',
    version=version,
    package_dir={'': 'src'},
    packages=[x for x in find_packages('src') if 'boogie.experimental' not in x],
    setup_requires='setuptools >= 30.3',
)



from pprint import pprint

from sidekick import fn

from .utils.text import safe_repr, first_line


#
# Printing
#
@fn
def info(obj, log=print, full=False):
    """
    Display detailed information about object.
    """

    log("=" * 60)
    log("Object:", obj)

    # Print object state
    try:
        dic = obj.__dict__
    except (AttributeError, TypeError):
        pass
    else:
        log("\nDict")
        if full:
            pprint(dic)
        else:
            _log_values(log, dic)

    # Print information about methods
    methods = []
    for attr, method in public_methods(obj):
        doc = method.__doc__ or ""
        if doc:
            methods.append(f"* {attr}: {first_line(doc)}")
        else:
            methods.append(f"* {attr}")
    if methods:
        log("\nMethods:")
        print("\n".join(methods))
    log("-" * 60)


def _log_values(log, values):
    # First print public values
    for k, v in values.items():
        if not k.startswith("_"):
            log(f"* {k}:", safe_repr(v, max_length=50))

    # Then the private ones (only types)
    for k, v in values.items():
        if k.startswith("_"):
            log(f"* {k} ({v.__class__.__name__})")


#
# Tracing
#
def set_trace():
    raise NotImplementedError


def embed():
    raise NotImplementedError


#
# Integration
#
def namespace():
    """
    Return a namespace of useful debug functions.
    """
    import boogie.debug as dbg

    return {
        "info": info,
        "log": print,
        "pprint": pprint,
        "set_trace": set_trace,
        "embed": embed,
        "dbg": dbg,
    }


def enable_debugging(auto=False):
    """
    Register debugging function to builtins.
    """
    import builtins

    # Checks if debugging should be enabled.
    if auto:
        pass

    # Configure environment in "debug" mode.
    for k, v in namespace().items():
        if not hasattr(builtins, k):
            setattr(builtins, k, v)


#
# Utility
#
def public_dir(obj):
    return (attr for attr in dir(obj) if not attr.startswith("_"))


def public_vars(obj):
    for attr in public_dir(obj):
        try:
            yield attr, getattr(obj, attr)
        except AttributeError:
            pass


def public_methods(obj):
    for attr, method in public_vars(obj):
        if callable(method):
            yield attr, method



from ..django_conf import Conf


class CompressorConf(Conf):
    # Django compressor
    COMPRESS_ENABLED = True
    COMPRESS_CSS_FILTERS = [
        "compressor.filters.css_default.CssAbsoluteFilter",
        "compressor.filters.cssmin.rCSSMinFilter",
    ]
    COMPRESS_PRECOMPILERS = [
        ("text/coffeescript", "coffee --compile --stdio"),
        ("text/less", "lessc {infile} {outfile}"),
        ("text/x-sass", "sass {infile} {outfile}"),
        ("text/x-scss", "sass --scss {infile} {outfile}"),
        ("text/stylus", "stylus < {infile} > {outfile}"),
    ]



from .environment import EnvironmentConf
from ..tools import secret_hash


class SecurityConf(EnvironmentConf):
    """
    Security options.
    """

    def finalize(self, settings):
        settings = super().finalize(settings)
        if not settings.get("SECRET_KEY"):
            settings["SECRET_KEY"] = secret_hash(settings)
        return settings

    def get_secret_key(self):
        """
        WARNING: keep the secret key used in production secret! We generate a
        secret from a hash of the current settings during the .finalize() phase.
        this is ok for local development, but may be insecure/inconvenient for
        """
        value = self.env.str("DJANGO_SECRET_KEY", default=None)
        if not value:
            if self.ENVIRONMENT in ("local", "test"):
                return self.ENVIRONMENT
            else:
                return None
        return value

    # https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators
    def get_auth_password_validators(self):
        """
        Password validation
        """
        prefix = "django.contrib.auth.password_validation"
        validators = [
            "UserAttributeSimilarityValidator",
            "MinimumLengthValidator",
            "CommonPasswordValidator",
            "NumericPasswordValidator",
        ]
        return [{"NAME": f"{prefix}.{x}"} for x in validators]

    def get_allowed_hosts(self):
        return self.env("DJANGO_ALLOWED_HOSTS", type=list, default=["localhost"])

    def get_password_hashers(self):
        if self.ENVIRONMENT == "testing":
            return ["django.contrib.auth.hashers.MD5PasswordHasher"]
        return [
            "django.contrib.auth.hashers.Argon2PasswordHasher",
            "django.contrib.auth.hashers.PBKDF2PasswordHasher",
            "django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher",
            "django.contrib.auth.hashers.BCryptSHA256PasswordHasher",
            "django.contrib.auth.hashers.BCryptPasswordHasher",
        ]



import copy
from itertools import chain

import pytest
from _pytest.fixtures import FixtureLookupError
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.http import Http404
from django.urls import resolve

from .client import Client
from .crawler import check_link_errors
from .urlchecker import LOGIN_REGEX, UrlChecker as _UrlChecker
from ..utils.text import snake_case

User = get_user_model()


#
# Fixture classes
#
class UserFixtures:
    """
    Create the anonymous, user, author and admin fixtures:

    anonymous:
        An anonymous user.
    user:
        Represents regular user with no special privileges on the platform.
    author:
        Another regular user, but we assume it to be the owner of some
        specific resource. A test scenario may create an ``user``, an ``author``
        and some page with ``page.owner = author``
    admin:
        A user with superuser permissions.

    For each user, it creates a corresponding client_<user> fixture that
    initializes a client logged in with the given user.
    """

    @pytest.fixture
    def anonymous(self, db):
        return AnonymousUser()

    @pytest.fixture
    def user(self, db):
        return self.make_user("user", email="user@user.com")

    @pytest.fixture
    def author(self, db):
        return self.make_user("author", email="author@author.com")

    @pytest.fixture
    def admin(self, db):
        return self.make_user(
            "admin", email="admin@admin.com", is_superuser=True, is_staff=True
        )

    @pytest.fixture
    def client(self, db):
        """Standard test client"""
        return Client()

    @pytest.fixture
    def user_client(self, db, client, user):
        """Client logged in as 'user'"""
        return self._with_login(client, user)

    @pytest.fixture
    def author_client(self, db, client, author):
        """Client logged in as 'author'"""
        return self._with_login(client, author)

    @pytest.fixture
    def admin_client(self, db, client, admin):
        """Client logged in as 'admin'"""
        assert admin.is_staff, f"User {admin} does not have staff privileges"
        assert admin.is_superuser, f"User {admin} does not have superuser privileges"
        return self._with_login(client, admin)

    def _with_login(self, client, user):
        # We create a copy because we don't want to share state between the
        # different client fixtures.
        client = copy.copy(client)
        client.force_login(user)
        return client

    def make_user(self, name, email, is_superuser=False, is_staff=False, **kwargs):
        """
        Creates a new user.

        Subclasses may override this function and it must accept the following
        arguments.

        Args:
            name:
                Fixture name for the user.
            email:
                User's e-mail.
            is_staff, is_superuser (bool):
                Tells if user can access the admin interface (staff) or is a
                sysadmin (superuser).
        """
        model = get_user_model()
        kwargs.update(is_superuser=is_superuser, is_staff=is_staff, email=email)
        return model.objects.create_user(name, **kwargs)


#
# URL Tester
#
class UrlTester(UserFixtures):
    """
    Test if users can access all paths in the "paths" attribute with a
    successful HTTP status code.

    Most of the logic for this class is implemented by
    :class:`boogie.testing.urlchecker.URLChecker` class.

    Examples:

        A UrlTester subclass must define the following (optional) attributes:

        .. code-block:: python

            class TestSomeAppUrls(UrlTester):
                # "paths" is a dictionary describing which paths can be
                # accessed by each user. Remember always using absolute paths.
                paths = {
                    # Use None to specify URLs that can be accessed by
                    # anonymous users.
                    None: [
                        '/some-app/',
                        '/some-app/url-a/',
                        '/some-app/url-b.json',
                    ],

                    # Other users are specified by strings which, during
                    # testing, are interpreted as fixture names.
                    'user': [
                        '/some-app/private/',
                        '/some-app/private-2/',
                    ],

                    # The list of paths is cumulative, i.e., "admin" inherits
                    # all URLs from "user", which inherits all URLs from None.
                    'admin': [
                        '/some-app/admin-panel/,
                        '/some-app/super-secret/,
                    ],
                }

                # The post_paths attr behaves similarly to paths, but specify
                # some payload data which is sent with a POST request.
                post_paths = {
                    None: {
                        '/some-app/form/: {
                            'name': 'Someone',
                            'email': 'some@email.com',
                        },
                    },

                    # Like paths, it re-uses entries associated with the
                    # previous fixtures. However, if a URL is repeated,
                    # it overrides the POST data.
                    'user': {
                        '/some-app/form/: {
                            'name': 'user',
                            'email': 'user@user.com',
                        },
                    },

                    # Append some unique #anchor to the path string to re-use
                    # the path of a previous user and include a new test data.
                    'admin': {
                        '/some-app/form/#admin: {
                            'name': 'admin',
                            'email': 'user@user.com',
                        },
                    },
                }

                # The URL tester will load a class scoped "data" fixture, if
                # available. This can be used to populate the database with
                # additional models necessary for the test to complete.
                @pytest.fixture
                def data(self, db):
                    models.Model1.objects.create(...)
                    models.Model2.objects.create(...)


                # Each user corresponds to a fixture in the database. We
                # automatically try to create a "user", "author" and "admin"
                # users. Of course you may want/need to personalize the creation
                # of those fixtures if you want.
                @pytest.fixture
                def user(self, db):
                    # The default user is created with the following command.
                    # The other fixtures follow a similar pattern.
                    return self.make_user('user', email='user@user.com')
    """

    paths: dict = {}
    post_paths: dict = {}
    login_regex: str = LOGIN_REGEX
    url_checker = _UrlChecker

    @pytest.fixture
    def data(self):
        return None

    @pytest.fixture
    def client(self, db):
        return Client()

    @pytest.mark.django_db
    def test_urls(self, request, client, data):
        """
        Implements the logic used by the dynamic test_urls class.
        """
        users = self.get_user_fixtures(request)
        kwargs = {"login_regex": self.login_regex, "client": client}
        checker = self.url_checker(self.paths, self.post_paths, **kwargs)
        errors = checker.check_url_errors(users=users)
        if errors:
            for url, value in errors.items():
                code = value.status_code
                print(f"Error fetching {url}, invalid response: {code}")
            raise AssertionError(f"errors found: {sorted(errors)}")

    def get_user_fixtures(self, request):
        """
        Creates a dictionary mapping names of user fixtures to their respective
        values.
        """
        users = set(chain(self.paths, self.post_paths))
        users.discard(None)
        return {user: request.getfixturevalue(user) for user in users}


#
# Web crawling
#
class CrawlerTester:
    """
    Base class for tests based on crawling your website from a initial URL.

    Can be configured by overriding the following attributes:

        start (path or list of paths):
            A url or a list of urls used as the starting points for the crawler.
        skip (list of urls):
            URLs that should be skipped during testing. The crawler will never
            visit those URLs unless they are a starting url in bases.
        xfail (list of urls):
            URLs that are expected to have failures.
        must_visit (list of urls):
            Checks if all URLs in that list were visited after crawling ends.
        user (str):
            Name of user instance fixture.
    """

    start = "/"
    user = "user"
    skip_patterns = ()
    skip_urls = ()
    xfail = ()
    must_visit = ()
    log = print
    error_class = AssertionError

    @pytest.fixture
    def conf(self, request):
        return {"user": request.getfixturevalue("user")}

    @pytest.mark.django_db
    def test_reaches_all_urls(self, request, conf):
        """
        Test if it fails in some view when crawling from a starting URL.
        """
        try:
            request.getfixturevalue("data")
        except FixtureLookupError:
            pass

        errors = {}
        user = conf["user"]
        start = self.start
        if isinstance(start, str):
            start = [start]

        for url in start:
            try:
                check_link_errors(
                    url,
                    visit=self.must_visit,
                    errors=self.xfail,
                    user=user,
                    skip_patterns=self.skip_patterns,
                    skip_urls=self.skip_urls,
                    log=self.log,
                )
            except AssertionError as ex:
                errors.update(ex.args[0])
        if errors:
            raise self.error_class(errors)


#
# Testing models
#
class ModelTester:
    """
    Perform some standard test on a given model

    This class assumes the presence of an "instance" fixture that creates an
    exemplary of the model class.

    Attributes:
        model (model class):
            The model class that to be tested.
        representation (str):
            Expected value of str(instance)
        absolute_url:
            Expected value of instance.get_absolute_url()
    """

    model: type
    representation: str
    absolute_url: str
    db = False

    @pytest.fixture
    def instance(self, request):
        model_name = snake_case(self.model.__name__)
        name = getattr(self, "instance_fixture", model_name)
        if self.db:
            request.getfixturevalue("db")
        try:
            return request.getfixturevalue(name)
        except FixtureLookupError:
            return self.get_instance()

    def get_instance(self):
        """
        Fallback method that is executed when the instance or model fixtures are
        not defined.
        """
        model_name = snake_case(self.model.__name__)
        raise ImproperlyConfigured(
            f'please either define an "instance" or "{model_name}"\n'
            f'fixture in your class or implement the "get_instance" method'
        )

    def get_examples(self):
        """
        Return a dictionary with some exemplary models with expected
        characteristics. Sub-classes can define any one of the entries bellow::

        - clean: passes a .full_clean() check
        - invalid: raises ValidationError during .full_clean()
        """
        return {}

    def test_django_model_interface(self, instance):
        """
        Test if model implements a basic Django Model interface correctly.
        """
        model = self.model
        self.check_improperly_configured(instance)

        # Check representation
        got = str(instance)
        expect = self.representation
        if expect != got:
            msg = "invalid representation: expect: %r, got: %r"
            raise AssertionError(msg % (expect, got))

            # Check absolute url
        if hasattr(model, "get_absolute_url"):
            assert instance.get_absolute_url() == self.absolute_url
            try:
                resolve(self.absolute_url)
            except Http404:
                raise AssertionError("absolute_url resulted on a 404")

        # Run examples
        for name, example in self.get_examples().items():
            if name == "clean":
                example.full_clean()
            elif example == "invalid":
                with pytest.raises(ValidationError):
                    example.full_clean()
            else:
                raise ImproperlyConfigured(f"invalid example: {name}")

    def check_improperly_configured(self, instance):
        """
        Check if test class was correctly set up for instance.
        """
        model = self.model
        requires_url = hasattr(model, "get_absolute_url")

        # Check success
        if hasattr(self, "representation") and (
            hasattr(self, "absolute_url") or not requires_url
        ):
            return

        print(
            "HINT: You can create a valid tester class by replacing it by\n"
            "the following code:"
        )
        print(make_test_class(self, instance))
        print("\nPlease revise to see if inferred properties are correct.")
        raise ImproperlyConfigured("Model tester class is not configured")


#
# Utility
#


def make_test_class(model_tester, instance):
    """
    Return a string with a plausible source code representation for a
    ModelTester test class.
    """

    base_names = ", ".join(cls.__name__ for cls in type(model_tester).__bases__)
    base = f"""
    class {model_tester.__class__.__name__}({base_names}):
        model = models.{instance.__class__.__name__}
        representation = {repr(str(instance))}"""
    if model_tester.db:
        base += "\n        db = True"
    if hasattr(instance, "get_absolute_url"):
        base += "\n        absolute_url = {instance.get_absolute_url()!r}"
    return base



from sidekick import import_later

# Function importer
_function = lambda x: import_later(x, package=__package__)

# Lazy imports of checker functions
check_link_errors = _function(".crawler:check_link_errors")
factory = _function(".factories:factory")



from django.core.exceptions import ImproperlyConfigured

try:
    import rest_framework as _rest
except ImportError:
    raise ImproperlyConfigured(
        "boogie.rest requires Django Rest Framework and some additional "
        "packages to be installed. Please install it using pip install "
        "django-boogie[rest]."
    )

from .rest_api import RestAPI
from .utils import to_json_default
from . import utils as _utils

rest_api = RestAPI()
_utils.patch_rest_framework_json_encoder()
